#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <devctl.h>
#include <hw/i2c.h>

#define NVRAM_I2C_ADDRESS 0x57

static int twifd;
static int nvram_init(void);
static int nvram_poke8(int twifd, unsigned short addr, unsigned char value);
static int nvram_peek8(int twifd, unsigned short addr);

static void usage(char **argv) {
	fprintf(stderr, "Usage: %s [OPTION] ...\n"
	  "Technologic Systems RTC nvram access\n"
	  "\n"
	  "  -h, --help            This message\n"
	  "  -a, --addr=<address>  Set the NVRAM address (0-127)\n"
	  "  -g, --get             Get the specified address\n"
	  "  -s, --set=<value>     Set the specified address to value\n\n",
	  argv[0]
	);
}


int main(int argc, char **argv)
{
	int twifd, c;
	int addr = -1;
	unsigned char data;
	

	if(argc == 1) {
		usage(argv);
		return(1);
	}
	twifd = nvram_init();
	if(twifd == -1)
		return 1;

	while((c = getopt(argc, argv, "ha:gs:")) != -1) {
		switch (c) {
		case 'a':
			addr = (unsigned char)strtoull(optarg, NULL, 0);
			if(addr > 127 || addr < 0){
				fprintf(stderr, "Address is out of bounds.  Must be in 0-127.\n");
				goto failed;
			}
			break;
		case 'g':
			if(addr == -1) {
				fprintf(stderr, "Address must be passed first\n");
				return 1;
			}
			data = nvram_peek8(twifd, addr);
			printf("nvram%d=0x%X\n", addr, data);
			break;
		case 's':
			if(addr == -1) {
				fprintf(stderr, "Address must be passed first\n");
				return 1;
			}
			data = (unsigned char)strtoull(optarg, NULL, 0);
			nvram_poke8(twifd, addr, data);
			break;
		case 'h':
		default:
			usage(argv);
		}
	}

	
	close(twifd);
	return 0;

	failed:
	close(twifd);
	return 1;
		
}


static int nvram_init(void)
{
	if(twifd != 0)
		return twifd;

	twifd = open("/dev/i2c1", O_RDWR);
   
	return twifd;
}


static int nvram_peek8(int twifd, unsigned short addr)
{
    iov_t           siov[2], riov[2];
    i2c_sendrecv_t  hdr;
    unsigned char reg;
    unsigned char val;

    reg = addr; 
    
    hdr.slave.addr = NVRAM_I2C_ADDRESS;
    hdr.slave.fmt = I2C_ADDRFMT_7BIT;
    hdr.send_len = 1;
    hdr.recv_len = 1;
    hdr.stop = 1;

    SETIOV(&siov[0], &hdr, sizeof(hdr));
    SETIOV(&siov[1], &reg, sizeof(reg));

    SETIOV(&riov[0], &hdr, sizeof(hdr));
    SETIOV(&riov[1], &val, sizeof(val));

    if (devctlv(twifd, DCMD_I2C_SENDRECV, 2, 2, siov, riov, NULL))
      return -1;
    else
      return val;
}

static int nvram_poke8(int twifd, unsigned short addr, unsigned char value)
{
    iov_t           siov[3];
    i2c_send_t      hdr;
    unsigned char reg;
    
    reg = addr;
    
    hdr.slave.addr = NVRAM_I2C_ADDRESS;
    hdr.slave.fmt = I2C_ADDRFMT_7BIT;
    hdr.len = 2;   
    hdr.stop = 1;

    SETIOV(&siov[0], &hdr, sizeof(hdr));
    SETIOV(&siov[1], &reg, sizeof(reg));
    SETIOV(&siov[2], &value, sizeof(value));

    return devctlv(twifd, DCMD_I2C_SEND, 3, 0, siov, NULL, NULL);
}

