/*
 * $QNXLicenseC: 
 * Copyright 2007, 2008, QNX Software Systems.  
 *  
 * Licensed under the Apache License, Version 2.0 (the "License"). You  
 * may not reproduce, modify or distribute this software except in  
 * compliance with the License. You may obtain a copy of the License  
 * at: http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" basis,  
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied. 
 * 
 * This file may contain contributions from others, either as  
 * contributors under the License or as licensors under other terms.   
 * Please review this entire file for other proprietary rights or license  
 * notices, as well as the QNX Development Suite License Guide at  
 * http://licensing.qnx.com/license-guide/ for other information. 
 * $ 
 */

#include "rtc.h"
#include <time.h>
#include <fcntl.h>
#include <hw/i2c.h>

/* 
 * Intersil ISL12020 Serial RTC
 */
#define ISL12020_SC          0   /* 00-59 */
#define ISL12020_MN          1   /* 00-59 */
#define ISL12020_HR          2   /* 0-1/00-23 */
#define ISL12020_DT          3   /* 01-31 */
#define ISL12020_MO          4   /* 01-12 */
#define ISL12020_YR          5   /* 00-99 */
#define ISL12020_DW          6   /* 00-06 */

#define ISL12020_INT         8   /* Interrupt Control Reg */


#define ISL12020I2C_ADDRESS  (0xDE >> 1)
#define ISL12020I2C_DEVNAME  "/dev/i2c1"


static int fd = -1;

static int
ISL12020i2c_read(unsigned char reg, unsigned char val[], unsigned char num)
{
    iov_t           siov[2], riov[2];
    i2c_sendrecv_t  hdr;

    hdr.slave.addr = ISL12020I2C_ADDRESS;
    hdr.slave.fmt = I2C_ADDRFMT_7BIT;
    hdr.send_len = 1;
    hdr.recv_len = num;
    hdr.stop = 1;

    SETIOV(&siov[0], &hdr, sizeof(hdr));
    SETIOV(&siov[1], &reg, sizeof(reg));

    SETIOV(&riov[0], &hdr, sizeof(hdr));
    SETIOV(&riov[1], val, num);
        
    return devctlv(fd, DCMD_I2C_SENDRECV, 2, 2, siov, riov, NULL);
}

static int
ISL12020i2c_write(unsigned char reg, unsigned char val[], unsigned char num)
{
    iov_t           siov[3];
    i2c_send_t      hdr;

    hdr.slave.addr = ISL12020I2C_ADDRESS;
    hdr.slave.fmt = I2C_ADDRFMT_7BIT;
    hdr.len = num + 1;
    hdr.stop = 1;

    SETIOV(&siov[0], &hdr, sizeof(hdr));
    SETIOV(&siov[1], &reg, sizeof(reg));
    SETIOV(&siov[2], val, num);

    return devctlv(fd, DCMD_I2C_SEND, 3, 0, siov, NULL, NULL);
}

int
RTCFUNC(init,isl12020)(struct chip_loc *chip, char *argv[])
{
    fd = open((argv && argv[0] && argv[0][0])? 
            argv[0]: ISL12020I2C_DEVNAME, O_RDWR);
    if (fd < 0) {
        fprintf(stderr, "Unable to open I2C device\n");
        return -1;
    }
    return 0;
}

int                         
RTCFUNC(get,isl12020)(struct tm *tm, int cent_reg)
{
    unsigned char   date[7];

    ISL12020i2c_read(ISL12020_SC, date, 7);

    tm->tm_sec  = BCD2BIN(date[ISL12020_SC] & 0x7f);
    tm->tm_min  = BCD2BIN(date[ISL12020_MN] & 0x7f);

	if ((date[ISL12020_HR] & 0x80) == 0) { /* MIL bit not set */
		/* the rtc is in 12 hour mode */ 
		int hour = BCD2BIN(date[ISL12020_HR] & 0x1f);

		if ((date[ISL12020_HR] & 0x20))  
			tm->tm_hour = (hour == 12) ? 12 : hour + 12; /* pm */
		else
			tm->tm_hour = (hour == 12) ? 0 : hour;  	 /* am */

	} else { 
		/* rejoice! the rtc is in 24 hour mode */ 
     	tm->tm_hour = BCD2BIN(date[ISL12020_HR] & 0x3f);
	}

    tm->tm_mday = BCD2BIN(date[ISL12020_DT] & 0x3f);
    tm->tm_mon  = BCD2BIN(date[ISL12020_MO] & 0x1f) - 1;
    tm->tm_year = BCD2BIN(date[ISL12020_YR] & 0xff) + 100;
	
    tm->tm_wday = BCD2BIN(date[ISL12020_DW] & 0x7);
    
    return(0);
}

int
RTCFUNC(set,isl12020)(struct tm *tm, int cent_reg)
{
    unsigned char   date[7];
    unsigned char reg;
	/* 
	 * Note: this function will set the clock incorrectly after 2099 
	 * And it sets the clock in 24 hour mode 
	 */
	 		 	 
	 if (tm->tm_year < 100) {
	    fprintf(stderr, "Error:  isl12020 RTC only good for years >= 2000\n");
	    return -1;
	 }
	 
    date[ISL12020_SC]   = BIN2BCD(tm->tm_sec); 
    date[ISL12020_MN]   = BIN2BCD(tm->tm_min);
    date[ISL12020_HR]   = BIN2BCD(tm->tm_hour) | 0x80; /* always set 24hr mode */
    date[ISL12020_DT]   = BIN2BCD(tm->tm_mday);
    date[ISL12020_MO]   = BIN2BCD(tm->tm_mon + 1);
    date[ISL12020_YR]   = BIN2BCD(tm->tm_year % 100);
    date[ISL12020_DW]   = BIN2BCD(tm->tm_wday);

	 reg = ISL12020i2c_read(ISL12020_INT, &reg, 1);
	 reg |= (1<<6);   /* set the WRTC bit */
	 ISL12020i2c_write(ISL12020_INT, &reg, 1);
 	 
    ISL12020i2c_write(ISL12020_SC, date, 7);

    return(0);
}

#if defined(__QNXNTO__) && defined(__USESRCVERSION)
#include <sys/srcversion.h>
__SRCVERSION("$URL: http://svn/product/branches/6.6.0/trunk/utils/r/rtc/clk_isl12020.c $ $Rev: 680331 $")
#endif
